==> yonderbox-graphql-mongodb-adapter-1.0.48.tgz: package/lib/adapterCacheInvalidation.js <==
const client = require('prom-client')
const pkg = require('../package.json')
const _ = require('lodash')

const yves = require('yves')

const debugCacheInvalidation = yves.debugger(`${pkg.name.replace(/-/g,':')}:cache:invalidation`)

const util = require('./util')

const cacheInvalidationTTL = ((1000 * 60) * 3) // 3 minutes
const cacheInvalidationCleanupInterval = 5 * 1000

const { MongoClient } = require('mongodb')

let AdapterDataSources
let mongoClient

const config = {
  enabled: true,
  url: 'mongodb://localhost:3001',
  db: 'adapter',
  collection: 'hits',
  options: {
    keepAlive: 1,
    connectTimeoutMS: 30000,
    reconnectTries: 30,
    reconnectInterval: 1000,
  },
}


function initialize() {
  if (!mongoClient) {
    connectMongodb(config).then( () => {

      mongoClient.db(config.db).createIndex(config.collection,{timestamp:-1})
      mongoClient.db(config.db).createIndex(config.collection,{markers:-1})

      let cursor = mongoClient.watch([])
      cursor.stream().on('data', doc => {

        const matchingDatasource = Object.keys(AdapterDataSources).filter( adapterName => doc.ns.db === _.get(AdapterDataSources[adapterName],`options.db`) )
        const ds = matchingDatasource.length ? AdapterDataSources[matchingDatasource[0]] : null

        if ( ds ) {
          if (ds.schema && ds.schema.collections[doc.ns.coll]) {
/*            debugCacheInvalidation('matches: %y %y',doc.ns.coll,ds.schema && ds.schema.collections[doc.ns.coll])*/
            debugCacheInvalidation('change stream %y',{
              db:doc.ns.db,
              collection:doc.ns.coll,
              operation:doc.operationType,
              id:doc.documentKey._id
            })
            if (doc.operationType != 'insert') {
              const marker = `${doc.ns.db}:${doc.ns.coll}:${doc.documentKey._id}`
              mongoClient.db(config.db)
                .collection(config.collection)
                .find({markers: marker})
                .toArray()
                .then((data) => {
                  if (data) {
                    data.forEach( hit => {
                      debugCacheInvalidation('hit: marker %y url %y',marker,hit._id)
                    })
                  }
                  return data
                }).catch( e => {
                  debug('exception: %y',e)
                }).finally( () => {} )

            }
          }
        }
      })

      debugCacheInvalidation('cleanup')
      function cleanup() {
        const stale = new Date( Date.now() - cacheInvalidationTTL )
        mongoClient.db(config.db)
          .collection(config.collection)
          .deleteMany({timestamp:{
            $lt: stale,
          }}).then( (response ) => {
            if (response.deletedCount) {
              debugCacheInvalidation('cleanup %y before %y',response.deletedCount, stale )
            }
            setTimeout( () => {
              cleanup()
            }, cacheInvalidationCleanupInterval)
          })
      }
      cleanup()
    })
  }
}

async function connectMongodb(options) {
  if (mongoClient) {
    return mongoClient
  } else {
    mongoClient = new MongoClient(options.url,Object.assign({ useNewUrlParser: true, useUnifiedTopology: true},options.options))
    return await mongoClient.connect().then( d => {
      debug('adapterCacheInvalidation MongoDB connected')
      return d
    }).catch( e => {
      debug('adapterCacheInvalidation MongoDB connect exceptions: %y',e)
    })
  }
}

function middleware (req, res, done) {
  if (req.headers['x-forwarded-for']) {
    res.on('finish', function () {
      const start = util.duration()
      const markers = _.get(res,'ctx.markers')
      if (markers && Object.keys(markers).length ) {
  /*      debugCacheInvalidation('markers %y',markers)
        debugCacheInvalidation('originalUrl %y ',req.originalUrl)
        debugCacheInvalidation('host %y ',req.host)
        debugCacheInvalidation('headers %y ',req.headers)
        debugCacheInvalidation('duration %y ms',util.duration(start))
  *//*      console.log(req)*/

        const fullMarkers=[]
        for (let adapter of Object.keys(markers)) {
  /*        debugCacheInvalidation('adapter %y',adapter)*/
          for (let collection of Object.keys(markers[adapter])) {
  /*          debugCacheInvalidation('collections %y',collection)*/
            for (let id of markers[adapter][collection]) {
  /*            debugCacheInvalidation('id %y',id)*/
              fullMarkers.push(`${adapter}:${collection}:${id}`)
            }
          }
        }
  /*      debugCacheInvalidation('fullMarkers %y',fullMarkers)*/

        const protocol = _.get(req,'headers.x-forwarded-proto','http')
        const url=`${protocol}://${req.headers.host}${req.originalUrl}`

        mongoClient.db(config.db)
          .collection(config.collection)
          .updateOne({_id:url},{$set:{headers:req.headers,markers:fullMarkers,timestamp:new Date()}},{upsert: true })
          .catch( err => {
            //          console.log(err)
            throw err
          })
          .finally( () => {
            debugCacheInvalidation('duration %y ms',util.duration(start))
          })

      }
    })
  }
  return done()
}

function instrument(server, adapterDataSources) {
  AdapterDataSources = adapterDataSources
  initialize()
  server.use(middleware)
}



module.exports = {
  instrument,
}

/*


const createSHA = require('crypto').createHash('sha256')

function computeQueryHash(query) {
    return createSHA.update(query).digest('hex')
}

*/

==> yonderbox-graphql-mongodb-adapter-1.0.50.tgz: package/lib/adapterCacheInvalidation.js <==
const pkg = require('../package.json')
const _ = require('lodash')

const yves = require('yves')

const debugCacheInvalidation = yves.debugger(`${pkg.name.replace(/-/g,':')}:cache:invalidation`)

const util = require('./util')

const cacheInvalidationTTL = ((1000 * 60) * 3) // 3 minutes
const cacheInvalidationCleanupInterval = 5 * 1000

const { MongoClient } = require('mongodb')

let AdapterDataSources
let mongoClient

const config = {
  enabled: true,
//  url: 'mongodb://localhost:3001',
  db: 'adapter',
  collection: 'hits',
  options: {
    keepAlive: 1,
    connectTimeoutMS: 30000,
//    reconnectTries: 30,
//    reconnectInterval: 1000,
  },
}

function setConfig(obj) {
  if (obj.url) config.url = obj.url
}

function initialize() {
  if (!mongoClient) {
    connectMongodb(config).then( () => {

      mongoClient.db(config.db).createIndex(config.collection,{timestamp:-1})
      mongoClient.db(config.db).createIndex(config.collection,{markers:-1})

      let cursor = mongoClient.watch([])
      cursor.stream().on('data', doc => {

        const matchingDatasource = Object.keys(AdapterDataSources).filter( adapterName => doc.ns.db === _.get(AdapterDataSources[adapterName],'options.db') )
        const ds = matchingDatasource.length ? AdapterDataSources[matchingDatasource[0]] : null

        if ( ds ) {
          if (ds.schema && ds.schema.collections[doc.ns.coll]) {
            /*            debugCacheInvalidation('matches: %y %y',doc.ns.coll,ds.schema && ds.schema.collections[doc.ns.coll])*/
            debugCacheInvalidation('change stream %y',{
              db:doc.ns.db,
              collection:doc.ns.coll,
              operation:doc.operationType,
              id:doc.documentKey._id
            })
            if (doc.operationType != 'insert') {
              const marker = `${doc.ns.db}:${doc.ns.coll}:${doc.documentKey._id}`
              mongoClient.db(config.db)
                .collection(config.collection)
                .find({markers: marker})
                .toArray()
                .then((data) => {
                  if (data) {
                    data.forEach( hit => {
                      debugCacheInvalidation('hit: marker %y url %y',marker,hit._id)
                    })
                  }
                  return data
                }).catch( e => {
                  debug('exception: %y',e)
                }).finally( () => {} )

            }
          }
        }
      })

      debugCacheInvalidation('cleanup')
      function cleanup() {
        const stale = new Date( Date.now() - cacheInvalidationTTL )
        mongoClient.db(config.db)
          .collection(config.collection)
          .deleteMany({timestamp:{
            $lt: stale,
          }}).then( (response ) => {
            if (response.deletedCount) {
              debugCacheInvalidation('cleanup %y before %y',response.deletedCount, stale )
            }
            setTimeout( () => {
              cleanup()
            }, cacheInvalidationCleanupInterval)
          })
      }
      cleanup()
    })
  }
}

async function connectMongodb(options) {
  if (mongoClient) {
    return mongoClient
  } else {
    mongoClient = new MongoClient(options.url,Object.assign({ useNewUrlParser: true, useUnifiedTopology: true},options.options))
    return await mongoClient.connect().then( d => {
      debug('adapterCacheInvalidation MongoDB connected')
      return d
    }).catch( e => {
      debug('adapterCacheInvalidation MongoDB connect exceptions: %y',e)
    })
  }
}

function middleware (req, res, done) {
  if (req.headers['x-forwarded-for']) {
    res.on('finish', function () {
      const start = util.duration()
      const markers = _.get(res,'ctx.markers')
      if (markers && Object.keys(markers).length ) {
        /*      debugCacheInvalidation('markers %y',markers)
        debugCacheInvalidation('originalUrl %y ',req.originalUrl)
        debugCacheInvalidation('host %y ',req.host)
        debugCacheInvalidation('headers %y ',req.headers)
        debugCacheInvalidation('duration %y ms',util.duration(start))
  *//*      console.log(req)*/

        const fullMarkers=[]
        for (let adapter of Object.keys(markers)) {
          /*        debugCacheInvalidation('adapter %y',adapter)*/
          for (let collection of Object.keys(markers[adapter])) {
            /*          debugCacheInvalidation('collections %y',collection)*/
            for (let id of markers[adapter][collection]) {
              /*            debugCacheInvalidation('id %y',id)*/
              fullMarkers.push(`${adapter}:${collection}:${id}`)
            }
          }
        }
        /*      debugCacheInvalidation('fullMarkers %y',fullMarkers)*/

        const protocol = _.get(req,'headers.x-forwarded-proto','http')
        const url=`${protocol}://${req.headers.host}${req.originalUrl}`

        mongoClient.db(config.db)
          .collection(config.collection)
          .updateOne({_id:url},{$set:{headers:req.headers,markers:fullMarkers,timestamp:new Date()}},{upsert: true })
          .catch( err => {
            //          console.log(err)
            throw err
          })
          .finally( () => {
            debugCacheInvalidation('duration %y ms',util.duration(start))
          })

      }
    })
  }
  return done()
}

function instrument(server, adapterDataSources) {
  AdapterDataSources = adapterDataSources
  initialize()
  server.use(middleware)
}



module.exports = {
  setConfig,
  instrument,
}

/*


const createSHA = require('crypto').createHash('sha256')

function computeQueryHash(query) {
    return createSHA.update(query).digest('hex')
}

*/
